import abc
from typing import Sequence

import numpy as np
import sympy as sp  # type: ignore

from scipy.stats import cauchy, laplace


class KoopmanObservable(abc.ABC):
    """
    Koopman Observables Functions
       These objects implement the mapping of the system state to the Koopman invariant space (explicitly).

    References
        Explanation of Koopman Observables
           Brunton, S. L., & Kutz, J. N. (2022). Data-driven science and engineering: Machine learning, dynamical systems,
           and control. Cambridge University Press. pp 260-261
    """

    @abc.abstractmethod
    def obs_fcn(self, X: np.ndarray) -> np.ndarray:
        """
        Observables Function

        :param X: system states
        :returns: observables
        """
        pass

    def __call__(self, X: np.ndarray) -> np.ndarray:
        return self.obs_fcn(X)

    def obs_grad(self, X: np.ndarray) -> np.ndarray:
        """
        Observables Gradient Function

        :param X: system states
        :returns: observables
        """
        raise NotImplementedError

    def __or__(self, obs: "KoopmanObservable"):
        """observable combination operator"""
        # TODO: implement combine
        return CombineObservable([self, obs])
        # identity | rffs


class CombineObservable(KoopmanObservable):
    def __init__(self, observables: Sequence[KoopmanObservable]):
        super(CombineObservable, self).__init__()
        self.observables = observables

    def obs_fcn(self, X: np.ndarray) -> np.ndarray:
        return np.vstack([obs.obs_fcn(X) for obs in self.observables])

    def obs_grad(self, X: np.ndarray) -> np.ndarray:
        return np.vstack([obs.obs_grad(X) for obs in self.observables])


class IdentityObservable(KoopmanObservable):
    def obs_fcn(self, X: np.ndarray) -> np.ndarray:
        return np.atleast_2d(X).T

    def obs_grad(self, X: np.ndarray) -> np.ndarray:
        assert len(X.shape) == 1
        return np.eye(len(X))


class SymbolicObservable(KoopmanObservable):
    def __init__(self, variables: Sequence[sp.Symbol], observables: Sequence[sp.Expr]):
        super(SymbolicObservable, self).__init__()
        self.length = len(variables)
        self._observables = observables
        self._variables = variables
        G = sp.Matrix(self._observables)
        GD = sp.Matrix([sp.diff(G, xi).T for xi in self._variables])
        self._g = sp.lambdify((self._variables,), G)
        self._gd = sp.lambdify((self._variables,), GD)

    def obs_fcn(self, X: np.ndarray) -> np.ndarray:
        return np.array(self._g(list(X.flatten())))

    def obs_grad(self, X: np.ndarray) -> np.ndarray:
        return np.array(self._gd(list(X))).T

    def __add__(self, obs: "SymbolicObservable"):
        return SymbolicObservable(
            list({*self._variables, *obs._variables}),
            [xi + yi for xi, yi in zip(self._observables, obs._observables)],
        )

    def __sub__(self, obs: "SymbolicObservable"):
        return SymbolicObservable(
            list({*self._variables, *obs._variables}),
            [xi - yi for xi, yi in zip(self._observables, obs._observables)],
        )

    def __mul__(self, obs: "SymbolicObservable"):
        return SymbolicObservable(
            list({*self._variables, *obs._variables}),
            [xi * yi for xi, yi in zip(self._observables, obs._observables)],
        )

    def __truediv__(self, obs: "SymbolicObservable"):
        return SymbolicObservable(
            list({*self._variables, *obs._variables}),
            [xi / yi for xi, yi in zip(self._observables, obs._observables)],
        )

    def __rdiv__(self, other):
        if isinstance(other, SymbolicObservable):
            return SymbolicObservable(
                list({*self._variables, *other._variables}),
                [xi / yi for xi, yi in zip(self._observables, other._observables)],
            )
        else:
            return SymbolicObservable(
                self._variables, [other / yi for yi in self._observables]
            )

    def __rmul__(self, other):
        if isinstance(other, SymbolicObservable):
            return SymbolicObservable(
                list({*self._variables, *other._variables}),
                [xi * yi for xi, yi in zip(self._observables, other._observables)],
            )
        else:
            return SymbolicObservable(
                self._variables, [other * yi for yi in self._observables]
            )

    def __or__(self, other):
        if isinstance(other, SymbolicObservable):
            return SymbolicObservable(
                list({*self._variables, *other._variables}),
                [*self._observables, *other._observables],
            )
        else:
            return CombineObservable([self, other])


class QuadraticObservable(SymbolicObservable):
    def __init__(self, length):
        """inefficient implementation to get quadratic koopman observables and its gradient functions"""
        vec = sp.symbols(" ".join([f"x{idx}" for idx in range(length)]))
        x = sp.Matrix((*vec, 1))
        U = x * x.T
        lv = [U[i, j] for i, j in zip(*np.tril_indices(len(x)))]
        super(QuadraticObservable, self).__init__(vec, lv)


class PolynomialObservable(KoopmanObservable):
    def __init__(self, dimension, degree) -> None:
        from sklearn.preprocessing import PolynomialFeatures

        super(PolynomialObservable, self).__init__()
        self.degree = degree
        self.dimension = dimension
        self.poly = PolynomialFeatures(int(self.degree), include_bias=False)
        self.poly.fit_transform(np.zeros((1, self.dimension)))

    def obs_fcn(self, X: np.ndarray) -> np.ndarray:
        return self.poly.transform(np.atleast_2d(X)).T


class RFFObservable(KoopmanObservable):
    def __init__(self, dimension, num_features, gamma, metric="rbf"):
        super(RFFObservable, self).__init__()
        self.gamma = gamma
        self.dimension = dimension
        self.metric = metric
        self.D = num_features
        # Generate D iid samples from p(w)
        if self.metric == "rbf":
            self.w = np.sqrt(2 * self.gamma) * np.random.normal(
                size=(self.D, self.dimension)
            )
        elif self.metric == "laplace":
            self.w = cauchy.rvs(scale=self.gamma, size=(self.D, self.dimension))
        # Generate D iid samples from Uniform(0,2*pi)
        self.u = 2 * np.pi * np.random.rand(self.D)

    def obs_fcn(self, X: np.ndarray) -> np.ndarray:
        # modification...
        if len(X.shape) == 1:
            x = np.atleast_2d(X.flatten()).T
        else:
            x = X.T
        w = self.w.T
        u = self.u[np.newaxis, :].T
        s = np.sqrt(2 / self.D)
        Z = s * np.cos(x.T @ w + u.T)
        return Z.T

    def obs_grad(self, X: np.ndarray) -> np.ndarray:
        if len(X.shape) == 1:
            x = np.atleast_2d(X.flatten()).T
        else:
            x = X.T
        x = np.atleast_2d(X.flatten()).T
        w = self.w.T
        u = self.u[np.newaxis, :].T
        s = np.sqrt(2 / self.D)
        # TODO: make this sparse?
        Z = -s * np.diag(np.sin(u + w.T @ x).flatten()) @ w.T
        return Z

    def compute_kernel(self, X: np.ndarray) -> np.ndarray:
        Z = self.transform(X)
        K = Z.dot(Z.T)
        return K



import autokoopman.core.system as asys
import sympy as sp


class PendulumWithInput(asys.SymbolicContinuousSystem):
    r"""
    Simple Pendulum with Constant Torque Input
        We model a pendulum with the equation:

        .. math::
            l^2 \ddot{\theta} + \beta \theta + g l \sin \theta = \tau

        This leads to the state space form:
        
        .. math::
            \left\{\begin{array}{l}
            \dot{x}_{1} = x_2 \\
            \dot{x}_{2} = - g / l \sin x_1 - 2 \beta x_2 + u_1 \\
            \end{array}\right.

        Note that :math:`\beta=b / m` kn this formulation: http://underactuated.mit.edu/pend.html .
    """

    def __init__(self, g=9.81, l=1.0, beta=0.0):
        theta, thetadot = sp.symbols("theta thetadot")
        tau = sp.symbols("tau")
        xdot = [thetadot, -g / l * sp.sin(theta) - 2 * beta * thetadot + tau]
        super(PendulumWithInput, self).__init__(
            (theta, thetadot), xdot, input_variables=(tau,)
        )



import pytest


def test_trajectory():
    import numpy as np
    import autokoopman.core.trajectory as traj

    # generate good and bad time state pairs
    states = np.random.random((10, 2))
    times = np.linspace(0.0, 2.0, 10)
    times_bad = np.linspace(0.0, 2.0, 8)

    # good state name
    names = [f"x{idx}" for idx in range(2)]

    # check that bad name fails
    with pytest.raises(AssertionError) as excinfo:
        traj = traj.Trajectory(times, states, states, ["wrong name"])

    # check that bad time (wrong number) fails
    with pytest.raises(AssertionError) as excinfo:
        traj_b = traj.Trajectory(times_bad, states, states)

    # good trajectory case
    traj = traj.Trajectory(times, states, None, names, None, threshold=0.01)
    assert len(traj.names) == 2

    # call some of the methods and see that they don't produce errors
    # TODO: test these better
    traj.interp1d(np.linspace(0.0, 2.0, 5))
    traj.norm()
    traj.interp_uniform_time(0.1)

    # check some properties
    assert traj.dimension == 2
    assert traj.states.shape[0] == 10
    assert traj.states.shape[1] == 2


def test_input_trajectory():
    import numpy as np
    import autokoopman.core.trajectory as traj

    # case: make good and bad state input pairs
    states = np.random.random((10, 2))
    inputs = np.random.random((10, 3))
    inputs_bad = np.random.random((7, 2))

    times = np.linspace(0.0, 2.0, 10)
    names_inputs = [f"x{idx}" for idx in range(3)]
    names = [f"x{idx}" for idx in range(2)]

    # case bad names
    with pytest.raises(AssertionError) as excinfo:
        traj_b = traj.Trajectory(times, states, inputs, names, ["wrong name"])

    with pytest.raises(AssertionError) as excinfo:
        traj_b = traj.Trajectory(times, states, inputs_bad, names, ["wrong name"])

    # create a good trajectory
    traj_g = traj.Trajectory(times, states, inputs, names, names_inputs)

    # call some of the methods and see that they don't produce errors
    # TODO: test these better
    traj_i = traj_g.interp1d(np.linspace(0.0, 2.0, 5))
    traj_g.norm()
    traj_g.interp_uniform_time(0.1)

    # check some properties
    for tr in {traj_i, traj_g}:
        assert tr.input_dimension == 3
        assert tr.dimension == 2
        assert tr.states.shape[0] in {5, 10}
        assert tr.states.shape[1] == 2



"""
Utilties to format strings
"""
import sys
import os
import warnings

from typing import Sequence, Optional


def _clip_str(s: Optional[str], nlength=20) -> str:
    """
    if a string is too long, clip it to a fixed length and add ...
    :param s: string
    :param nlength: length to clip
    :return: clipped string
    """
    if s is None:
        return "None"
    if len(s) <= nlength:
        return s
    else:
        return f"{s[:nlength]}..."


def _clip_list(s: Optional[Sequence], nlength=20) -> str:
    """
    if a list is too long as a string, clip it to a fixed length
    :param s: string
    :param nlength: length to clip
    :return: clipped string
    """
    if s is None:
        return "None"
    return f"[{', '.join([str(si) for si in s])}]"


class hide_prints:
    """context manager to suppress python print output

    this was created to hide "ugly stuff"
    (e.g., GPyOpt prints)
    """

    def __enter__(self, hide_warnings=True):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")
        self.hw = hide_warnings

        # TODO: it might make more sense to capture this
        # as a log
        if self.hw:
            self.cw = warnings.catch_warnings()
            self.cw.__enter__()
            warnings.filterwarnings("ignore")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout
        if self.hw:
            self.cw.__exit__()



import autokoopman.core.estimator as kest
import autokoopman.core.system as ksys
import numpy as np
from typing import Optional


def dmdc(X, Xp, U, r):
    """Dynamic Mode Decomposition with Control (DMDC)

    DMD but extended to include control.

    References
        Proctor, J. L., Brunton, S. L., & Kutz, J. N. (2016). Dynamic mode decomposition with control. SIAM Journal on Applied Dynamical Systems, 15(1), 142-161.

        See https://arxiv.org/pdf/1409.6358.pdf for more details
    """
    if U is not None:
        Y = np.hstack((X, U)).T
    else:
        Y = X.T
    Yp = Xp.T
    state_size = Yp.shape[0]

    # compute Atilde
    U, Sigma, V = np.linalg.svd(Y, False)
    U, Sigma, V = U[:, :r], np.diag(Sigma)[:r, :r], V.conj().T[:, :r]

    # get the transformation
    Atilde = Yp @ V @ np.linalg.inv(Sigma) @ U.conj().T
    return Atilde[:, :state_size], Atilde[:, state_size:]


class KoopmanDiscEstimator(kest.NextStepEstimator):
    """Koopman Discrete Estimator

    This methods implements a regularized form of Koopman with Inputs (KIC). It assumes that the input
    if piecewise constant, zeroing out some of the state + input transform.

    TODO: add other ways to implement KIC
    TODO: sampling period isn't used

    :param observables: function that returns the observables of the system state
    :param sampling_period: sampling period of the uniform time, discrete system
    :param dim: dimension of the system state
    :param rank: rank of the DMDc

    References
        Proctor, J. L., Brunton, S. L., & Kutz, J. N. (2018). Generalizing Koopman theory to allow for inputs and control. SIAM Journal on Applied Dynamical Systems, 17(1), 909-930.

        See https://epubs.siam.org/doi/pdf/10.1137/16M1062296 for more details
    """

    def __init__(self, observables, sampling_period, dim, rank, **kwargs):
        super().__init__(**kwargs)
        self.dim = dim
        self.obs = observables
        self.rank = int(rank)
        self._A, self._B = None, None

    def fit_next_step(
        self, X: np.ndarray, Y: np.ndarray, U: Optional[np.ndarray] = None
    ) -> None:
        """fits the discrete system model

        calls DMDC after building out the observables
        """
        G = np.array([self.obs(xi).flatten() for xi in X.T])
        Gp = np.array([self.obs(xi).flatten() for xi in Y.T])
        self._A, self._B = dmdc(G, Gp, U.T if U is not None else U, self.rank)
        self._has_input = U is not None

    @property
    def model(self) -> ksys.System:
        """
        packs the learned linear transform into a discrete linear system
        """
        return ksys.KoopmanStepDiscreteSystem(
            self._A, self._B, self.obs, self.names, self.dim, self.scaler
        )


class KoopmanContinuousEstimator(kest.GradientEstimator):
    """Koopman Continuous Estimator

    :param observables: function that returns the observables of the system state
    :param dim: dimension of the system state
    :param rank: rank of the DMDc

    References
        Proctor, J. L., Brunton, S. L., & Kutz, J. N. (2018). Generalizing Koopman theory to allow for inputs and control. SIAM Journal on Applied Dynamical Systems, 17(1), 909-930.

        See https://epubs.siam.org/doi/pdf/10.1137/16M1062296 for more details

    """

    def __init__(self, observables, dim, rank, **kwargs):
        super().__init__(**kwargs)
        self.dim = dim
        self.obs = observables
        self.rank = int(rank)
        self._A, self._B = None, None

    def fit_gradient(
        self, X: np.ndarray, Y: np.ndarray, U: Optional[np.ndarray] = None
    ) -> None:
        """fits the gradient system model

        calls DMDC after building out the observables
        """
        G = np.array([self.obs(xi).flatten() for xi in X.T])
        Gp = np.array([self.obs(xi).flatten() for xi in Y.T])
        self._A, self._B = dmdc(G, Gp, U.T if U is not None else U, self.rank)
        self._has_input = U is not None

    @property
    def model(self) -> ksys.System:
        """
        packs the learned linear transform into a continuous linear system
        """
        return ksys.KoopmanGradientContinuousSystem(
            self._A, self._B, self.obs, self.names, self.dim, self.scaler
        )



import autokoopman.core.tuner as atuner
from autokoopman.core.trajectory import TrajectoriesData
from autokoopman.core.tuner import TuneResults, TrajectoryScoring
from typing import Callable


class MonteCarloTuner(atuner.HyperparameterTuner):
    def tune(
        self,
        nattempts=100,
        scoring_func: Callable[
            [TrajectoriesData, TrajectoriesData], float
        ] = TrajectoryScoring.end_point_score,
    ) -> TuneResults:
        sampling = self.tune_sampling(nattempts, scoring_func)
        next(sampling)

        while True:
            param = self._parameter_model.parameter_space.random()
            try:
                sampling.send(param)
                next(sampling)
            except StopIteration:
                break
            except Exception as exc:
                print(f"Error: {exc}")
                self.error_messages.append((param, exc))
        return self.best_result



import setuptools
import versioneer

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="autokoopman",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Automated Koopman Operator Linearization Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EthanJamesLew/AutoKoopman",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering"
    ],
    package_dir={"": ".", "autokoopman": "autokoopman"},
    install_requires=[
        "numpy>=1.19.0",
        "matplotlib>=3.3.4",
        "scipy>=1.5.4",
        "sympy >= 1.9",
        "pandas >= 1.1.5",
        "scikit-learn >= 1.0.2",
        "tqdm>=4.62.2",
        "pysindy==1.7.2",
    ],
    extra_requires={
        'deepk': [
            "torch>=1.12.1",
        ],
        'bopt': [
            "gpy==1.10.0",
            "gpyopt==1.2.6",
        ],
    },
    packages=setuptools.find_packages(where="."),
    python_requires=">=3.8",
)



import sympy as sp  # type: ignore

import autokoopman.core.system as asys


class Bio2(asys.SymbolicContinuousSystem):
    r"""
    Spring pendulum
        A nine-dimensional continuous model which is adapted from a biological system given in [1].

        Initial set:
            x1 in [0.99,1.01]
            x2 in [0.99,1.01]
            x3 in [0.99,1.01]
            x4 in [0.99,1.01]
            x5 in [0.99,1.01]
            x6 in [0.99,1.01]
            x7 in [0.99,1.01]
            x8 in [0.99,1.01]
            x9 in [0.99,1.01]

    References
         E. Klipp, R. Herwig, A. Kowald, C. Wierling, H. Lehrach. Systems Biology in Practice: Concepts, Implementation and Application. Wiley-Blackwell, 2005.
    """

    def __init__(self):
        x1, x2, x3, x4, x5, x6, x7, x8, x9 = sp.symbols("x1 x2 x3 x4 x5 x6 x7 x8 x9   ")
        xdot = [
            3 * x3 - x1 * x6,
            x4 - x2 * x6,
            x1 * x6 - 3 * x3,
            x2 * x6 - x4,
            3 * x3 + 5 * x1 - x5,
            5 * x5 + 3 * x3 + x4 - x6 * (x1 + x2 + 2 * x8 + 1),
            5 * x4 + x2 - 0.5 * x7,
            5 * x7 - 2 * x6 * x8 + x9 - 0.2 * x8,
            2 * x6 * x8 - x9,
        ]
        super(Bio2, self).__init__((x1, x2, x3, x4, x5, x6, x7, x8, x9), xdot)
